import itertools as _itertools

import six as _six


def flatten(gen):
    """Flatten an iterable.

    >>> flat = flatten([(1, 2, 3), [4, 5, 6], range(7, 10)])
    >>> list(flat)
    [1, 2, 3, 4, 5, 6, 7, 8, 9]

    """
    return _itertools.chain.from_iterable(gen)


def recursive_flatten(elements):
    """Flatten an iterable.

    This implementation is slower than `flatten` but works recursively.

    >>> gen = recursive_flatten([1, [2, [3, [lambda blah: blah]]]])
    >>> l = list(gen)
    >>> l[:3]
    [1, 2, 3]
    >>> import types
    >>> assert type(l[-1]) == types.FunctionType

    """
    gen = iter(elements)
    for element in gen:
        if is_iterable(element) and not isinstance(element, _six.string_types):
            for element in recursive_flatten(element):
                yield element
        else:
            yield element


def is_iterable(testee):
    """Check if a value is iterable.

    This does no type comparisons. Note that strings are iterables too!

    >>> is_iterable([])
    True

    >>> is_iterable("Hello World!")
    True

    >>> is_iterable(False)
    False

    """
    try:
        iter(testee)
        return True
    except TypeError:
        return False


def chunk(chunk_size=32):
    """Chunk `chunk_size` elements of an iterable.

    Every yielded chunk will be a list. The last chunk may contain less
    elements than `chunk_size`.

    >>> chunker = chunk(32)
    >>> gen = chunker(range(64))
    >>> assert len(list(gen)) == 2

    >>> chunker = chunk(3)
    >>> list(chunker([1,2,3,4,5,6,7,8]))
    [[1, 2, 3], [4, 5, 6], [7, 8]]

    """
    if chunk_size < 1:
        raise ValueError("chunk_size can not be less than 1, got %d" %
                         chunk_size)
    key = lambda x, s=chunk_size: x[0] // s

    def chunker(gen):
        """ Actual chunker """
        for _, grouped_chunk in _itertools.groupby(enumerate(gen), key=key):
            yield [entry[1] for entry in grouped_chunk]

    return chunker



import resource
import time
import logging


class Counter(object):
    """
    >>> c = Counter()
    >>> c.init("reads")
    >>> c.init("writes")
    >>> c.count("reads")
    >>> c.count("writes")
    >>> c.show()

    """
    __slots__ = ('counts', 'start', 'maxlength', 'formatstring', 'last_show',
                 'session', 'show_every', 'items', 'shows', 'quiet', 'logger')

    def __init__(self, logger=None, session=None, show_every=10, quiet=False):
        self.session = session
        if logger and isinstance(logger, (unicode, str)):
            self.logger = logging.getLogger(logger)
        elif logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger("core.pipeline")

        self.quiet = bool(quiet)
        self.items = []
        self.counts = {}
        self.start = {}
        self.show_every = show_every
        self.formatstring = ""
        self.maxlength = 0
        self.last_show = None
        self.shows = 0

    def init(self, name):
        self.items.append(name)
        self.counts[name] = 0
        self.start[name] = time.time()
        self.maxlength = max(self.maxlength, len(name))
        self.formatstring = self.formatstring + " %d/s "
        self.last_show = time.time()

    def show(self, message=None):
        if self.quiet:
            return

        message = []
        now = time.time()

        if not self.shows:
            self.shows = 25
            widths = [14] * len(self.items) + [5, 3]
            headers = zip(self.items + ["Total", "Mem"],
                          widths)

            head_msg = " | ".join(("%%%ds" % width) % header
                                  for header, width in headers)
            self.logger.debug(head_msg)

        self.shows -= 1

        for name in self.items:
            count = self.counts[name]
            total = count
            message.append(
                "%6d %5d/s" % (count, (count / (now - self.start[name])))
            )

        message.append("%5d" % total)

        memory_used = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        message.append("[%s]" % memory_used)
        if self.session:
            message.append("[new:%d, dirty:%d, deleted:%d, id_map:%d]" % (
                len(self.session.new),
                len(self.session.dirty),
                len(self.session.deleted),
                len(self.session.identity_map),
            ))
        self.logger.debug(" | ".join(message))

    def gauge(self, name, value):
        if name not in self.items:
            self.init(name)
        self.counts[name] = value

    def count(self, name="default", count=1):
        if name not in self.items:
            self.init(name)

        now = time.time()
        self.counts[name] += count
        if self.show_every is None:
            return

        if self.show_every == -1 or (now - self.last_show > self.show_every):
            self.show()
            self.last_show = now



#!/usr/bin/env python
# encoding: utf-8
"""
Pipeline processing
"""
from .threaded import AbstractStep, Pipeline, SkipEntry, iterstep, funcstep

__all__ = ["AbstractStep", "Pipeline", "SkipEntry", "iterstep", "funcstep"]



import types as _types
import functools as _functools
import logging as _logging

import six as _six

from .counter import Counter
from .iterables import recursive_flatten


logger = _logging.getLogger(__name__)


class PipelineStep(object):

    def process(self, iterator):
        return iterator

    def __call__(self, iterator):
        return self.process(iterator)


def identity_step(iterator):
    """A pipeline-step which does nothing."""
    return iterator


def entry_counter(iterator):
    counter = Counter(show_every=1)
    counter.init("products")
    for entry in iterator:
        counter.count("products")
        yield entry
    counter.show()


def defer_call(step_function):
    """Proxy argument passing of iterators.

    This decorator passes the argument signature on down the
    iterator pipeline, and the result back up.

    """
    def wraps(callable_thing):
        if isinstance(callable_thing, _types.FunctionType):
            return _functools.wraps(callable_thing)
        else:
            assert _six.callable(callable_thing)
            return identity_step

    @wraps(step_function)
    def wrapper(iterator_factory):
        @wraps(iterator_factory)
        def wrap(*args, **kw):
            return step_function(iterator_factory(*args, **kw))
        return wrap
    return wrapper


def combine_pipeline(source, pipeline, debugger=None):
    """Build the "pipeline" generator.

    Optionally a debugging object may be passed in.

    >>> def step(iterator):
    ...     for entry in iterator:
    ...         yield entry

    >>> def remover(iterator):
    ...     for count, entry in enumerate(iterator):
    ...         if count % 2 == 0:
    ...             yield entry

    >>> class Identity(PipelineStep):
    ...     def process(self, iterator):
    ...         for entry in iterator:
    ...             yield entry
    >>> assert hasattr(Identity, '__call__')

    It's possible for iterator-factories to return lists of iterators. These
    will then be flattened into the pipeline in the order of occurence.

    >>> pipeline = [step, [remover, Identity()], step]

    Example: We construct 1000 elements and pass it through our pipeline. As
    the remover removes every second element we can expect to have only half
    as many elements out than we sent in.

    >>> gen = combine_pipeline(range(1000), pipeline)
    >>> assert len(list(gen)) == 500

    # FIXME: Add sample debugging implementation and test it

    >>> #gen = combine_pipeline(range(1000), pipeline, debug=True)
    >>> #assert len(list(gen)) == 500

    The `source` can also be a callable. If it is, `combine_pipeline` will also
    return a callable, which - when called - will pass the supplied arguments
    on to the source callable.

    >>> gen = combine_pipeline(range, pipeline)
    >>> assert len(list(gen(1000))) == 500

    >>> #gen = combine_pipeline(xrange, pipeline, debug=True)
    >>> #assert len(list(gen(1000))) == 500

    """
    identity = lambda x: x

    if debugger is not None:
        inst = debugger()
        head = inst.head
        tail = inst.tail
        track = inst.track
        report = inst.report
    else:
        head = tail = track = report = identity

    gen = source
    if _six.callable(gen):
        # Source is (hopefully) an iterator-factory, so we need
        # to convert all our steps into iterator-factories
        # as well, to enable passing on the arguments passed
        # to our collapsed pipeline.
        defer = defer_call
    else:
        defer = identity

    track = defer(track)
    head = defer(head)
    tail = defer(tail)
    report = defer(report)

    gen = head(gen)

    for step in recursive_flatten(pipeline):
        gen = defer(step)(track(gen))

    return report(tail(track(gen)))



from setuptools import setup

setup(
    name='paipa',
    version='0.0.2',
    description='Python pipeline library. Maori: (noun) pipe.',
    packages=['paipa'],
    tests_require=[
        'tox',
    ],
    install_requires=[
        'six',
        'enum34',
    ],
    extras_require={
        'glue': ['tornado'],
    },
)
